@@ -25,6 +25,7 @@
  */
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdint.h>
 #include <unistd.h>
 #include <string.h>
 #include <errno.h>
@@ -38,11 +39,6 @@
 #include <arpa/inet.h>
 #include <ev.h>
 #include <assert.h>
-
-#ifdef HAVE_ALLOCA_H
-#include <alloca.h>
-#endif
-
 #include "connection.h"
 #include "resolv.h"
 #include "address.h"
@@ -57,8 +53,9 @@
 
 struct resolv_cb_data {
     struct Connection *connection;
-    struct Address *address;
+    const struct Address *address;
     struct ev_loop *loop;
+    int cb_free_addr;
 };
 
 
@@ -74,6 +71,7 @@ static void reactivate_watcher(struct ev_loop *, struct ev_io *,
 static void connection_cb(struct ev_loop *, struct ev_io *, int);
 static void resolv_cb(struct Address *, void *);
 static void reactivate_watchers(struct Connection *, struct ev_loop *);
+static void insert_proxy_v1_header(struct Connection *);
 static void parse_client_request(struct Connection *);
 static void resolve_server_address(struct Connection *, struct ev_loop *);
 static void initiate_server_connect(struct Connection *, struct ev_loop *);
@@ -106,12 +104,13 @@ accept_connection(struct Listener *listener, struct ev_loop *loop) {
         err("new_connection failed");
         return 0;
     }
+    con->listener = listener_ref_get(listener);
 
 #ifdef HAVE_ACCEPT4
     int sockfd = accept4(listener->watcher.fd,
                     (struct sockaddr *)&con->client.addr,
                     &con->client.addr_len,
-                    O_NONBLOCK);
+                    SOCK_NONBLOCK);
 #else
     int sockfd = accept(listener->watcher.fd,
                     (struct sockaddr *)&con->client.addr,
@@ -132,18 +131,32 @@ accept_connection(struct Listener *listener, struct ev_loop *loop) {
     fcntl(sockfd, F_SETFL, flags | O_NONBLOCK);
 #endif
 
+    if (getsockname(sockfd, (struct sockaddr *)&con->client.local_addr,
+                &con->client.local_addr_len) != 0) {
+        int saved_errno = errno;
+
+        warn("getsockname failed: %s", strerror(errno));
+        free_connection(con);
+
+        errno = saved_errno;
+        return 0;
+    }
+
     /* Avoiding type-punned pointer warning */
     struct ev_io *client_watcher = &con->client.watcher;
     ev_io_init(client_watcher, connection_cb, sockfd, EV_READ);
     con->client.watcher.data = con;
     con->state = ACCEPTED;
-    con->listener = listener_ref_get(listener);
     con->established_timestamp = ev_now(loop);
 
     TAILQ_INSERT_HEAD(&connections, con, entries);
 
     ev_io_start(loop, client_watcher);
 
+    if (con->listener->table->use_proxy_header ||
+            con->listener->fallback_use_proxy_header)
+        insert_proxy_v1_header(con);
+
     return 1;
 }
 
@@ -227,6 +240,8 @@ static void
 connection_cb(struct ev_loop *loop, struct ev_io *w, int revents) {
     struct Connection *con = (struct Connection *)w->data;
     int is_client = &con->client.watcher == w;
+    const char *socket_name =
+        is_client ? "client" : "server";
     struct Buffer *input_buffer =
         is_client ? con->client.buffer : con->server.buffer;
     struct Buffer *output_buffer =
@@ -238,7 +253,8 @@ connection_cb(struct ev_loop *loop, struct ev_io *w, int revents) {
     if (revents & EV_READ && buffer_room(input_buffer)) {
         ssize_t bytes_received = buffer_recv(input_buffer, w->fd, 0, loop);
         if (bytes_received < 0 && !IS_TEMPORARY_SOCKERR(errno)) {
-            warn("recv(): %s, closing connection",
+            warn("recv(%s): %s, closing connection",
+                    socket_name,
                     strerror(errno));
 
             close_socket(con, loop);
@@ -253,7 +269,8 @@ connection_cb(struct ev_loop *loop, struct ev_io *w, int revents) {
     if (revents & EV_WRITE && buffer_len(output_buffer)) {
         ssize_t bytes_transmitted = buffer_send(output_buffer, w->fd, 0, loop);
         if (bytes_transmitted < 0 && !IS_TEMPORARY_SOCKERR(errno)) {
-            warn("send(): %s, closing connection",
+            warn("send(%s): %s, closing connection",
+                    socket_name,
                     strerror(errno));
 
             close_socket(con, loop);
@@ -342,12 +359,88 @@ reactivate_watcher(struct ev_loop *loop, struct ev_io *w,
     }
 }
 
+static void
+insert_proxy_v1_header(struct Connection *con) {
+    char buf[INET6_ADDRSTRLEN] = { '\0' };
+    size_t buf_len;
+
+    con->header_len += buffer_push(con->client.buffer, "PROXY ", 6);
+
+    switch (con->client.addr.ss_family) {
+        case AF_INET:
+            con->header_len += buffer_push(con->client.buffer, "TCP4 ", 5);
+
+            inet_ntop(AF_INET,
+                      &((const struct sockaddr_in *)&con->client.addr)->
+                      sin_addr, buf, sizeof(buf));
+            buf_len = strlen(buf);
+            con->header_len += buffer_push(con->client.buffer, buf, buf_len);
+
+            con->header_len += buffer_push(con->client.buffer, " ", 1);
+
+            inet_ntop(AF_INET,
+                      &((const struct sockaddr_in *)&con->client.local_addr)->
+                      sin_addr, buf, sizeof(buf));
+            buf_len = strlen(buf);
+            con->header_len += buffer_push(con->client.buffer, buf, buf_len);
+
+            buf_len = snprintf(buf, sizeof(buf), " %" PRIu16,
+                              ntohs(((const struct sockaddr_in *)&con->
+                              client.addr)->sin_port));
+            con->header_len += buffer_push(con->client.buffer, buf, buf_len);
+
+            buf_len = snprintf(buf, sizeof(buf), " %" PRIu16,
+                              ntohs(((const struct sockaddr_in *)&con->
+                              client.local_addr)->sin_port));
+            con->header_len += buffer_push(con->client.buffer, buf, buf_len);
+
+            break;
+        case AF_INET6:
+            con->header_len += buffer_push(con->client.buffer, "TCP6 ", 5);
+            inet_ntop(AF_INET6,
+                    &((const struct sockaddr_in6 *)&con->client.addr)->
+                    sin6_addr, buf, sizeof(buf));
+            buf_len = strlen(buf);
+            con->header_len += buffer_push(con->client.buffer, buf, buf_len);
+
+            con->header_len += buffer_push(con->client.buffer, " ", 1);
+
+            inet_ntop(AF_INET6,
+                      &((const struct sockaddr_in6 *)&con->
+                      client.local_addr)->sin6_addr, buf, sizeof(buf));
+            buf_len = strlen(buf);
+            con->header_len += buffer_push(con->client.buffer, buf, buf_len);
+
+            buf_len = snprintf(buf, sizeof(buf), " %" PRIu16,
+                              ntohs(((const struct sockaddr_in6 *)&con->
+                              client.addr)->sin6_port));
+            con->header_len += buffer_push(con->client.buffer, buf, buf_len);
+
+            buf_len = snprintf(buf, sizeof(buf), " %" PRIu16,
+                              ntohs(((const struct sockaddr_in6 *)&con->
+                              client.local_addr)->sin6_port));
+            con->header_len += buffer_push(con->client.buffer, buf, buf_len);
+
+            break;
+        default:
+            con->header_len += buffer_push(con->client.buffer, "UNKNOWN", 7);
+    }
+    con->header_len += buffer_push(con->client.buffer, "\r\n", 2);
+}
+
 static void
 parse_client_request(struct Connection *con) {
     const char *payload;
-    ssize_t payload_len = buffer_coalesce(con->client.buffer, (const void **)&payload);
+    size_t payload_len = buffer_coalesce(con->client.buffer, (const void **)&payload);
     char *hostname = NULL;
 
+    /* Avoid payload_len underflow and empty request */
+    if (payload_len <= con->header_len)
+        return;
+
+    payload += con->header_len;
+    payload_len -= con->header_len;
+
     int result = con->listener->protocol->parse_packet(payload, payload_len, &hostname);
     if (result < 0) {
         char client[INET6_ADDRSTRLEN + 8];
@@ -356,7 +449,7 @@ parse_client_request(struct Connection *con) {
             if (buffer_room(con->client.buffer) > 0)
                 return; /* give client a chance to send more data */
 
-            warn("Request from %s exceeded %ld byte buffer size",
+            warn("Request from %s exceeded %zu byte buffer size",
                     display_sockaddr(&con->client.addr, client, sizeof(client)),
                     buffer_size(con->client.buffer));
         } else if (result == -2) {
@@ -378,7 +471,7 @@ parse_client_request(struct Connection *con) {
     }
 
     con->hostname = hostname;
-    con->hostname_len = result;
+    con->hostname_len = (size_t)result;
     con->state = PARSED;
 }
 
@@ -395,30 +488,37 @@ abort_connection(struct Connection *con) {
 
 static void
 resolve_server_address(struct Connection *con, struct ev_loop *loop) {
-    /* TODO avoid extra malloc in listener_lookup_server_address() */
-    struct Address *server_address =
+    struct LookupResult result =
         listener_lookup_server_address(con->listener, con->hostname, con->hostname_len);
 
-    if (server_address == NULL) {
+    if (result.address == NULL) {
         abort_connection(con);
         return;
-    } else if (address_is_hostname(server_address)) {
+    } else if (address_is_hostname(result.address)) {
 #ifndef HAVE_LIBUDNS
         warn("DNS lookups not supported unless sniproxy compiled with libudns");
-        free(server_address);
+
+        if (result.caller_free_address)
+            free((void *)result.address);
+
         abort_connection(con);
         return;
 #else
         struct resolv_cb_data *cb_data = malloc(sizeof(struct resolv_cb_data));
         if (cb_data == NULL) {
             err("%s: malloc", __func__);
-            free(server_address);
+
+            if (result.caller_free_address)
+                free((void *)result.address);
+
             abort_connection(con);
             return;
         }
         cb_data->connection = con;
-        cb_data->address = server_address;
+        cb_data->address = result.address;
+        cb_data->cb_free_addr = result.caller_free_address;
         cb_data->loop = loop;
+        con->use_proxy_header = result.use_proxy_header;
 
         int resolv_mode = RESOLV_MODE_DEFAULT;
         if (con->listener->transparent_proxy) {
@@ -434,26 +534,28 @@ resolve_server_address(struct Connection *con, struct ev_loop *loop) {
                     warn("attempt to use transparent proxy with hostname %s "
                             "on non-IP listener %s, falling back to "
                             "non-transparent mode",
-                            address_hostname(server_address),
+                            address_hostname(result.address),
                             display_sockaddr(con->listener->address,
                                     listener_address, sizeof(listener_address))
                             );
             }
         }
 
-        con->query_handle = resolv_query(address_hostname(server_address),
+        con->query_handle = resolv_query(address_hostname(result.address),
                 resolv_mode, resolv_cb,
                 (void (*)(void *))free_resolv_cb_data, cb_data);
 
         con->state = RESOLVING;
 #endif
-    } else if (address_is_sockaddr(server_address)) {
-        con->server.addr_len = address_sa_len(server_address);
+    } else if (address_is_sockaddr(result.address)) {
+        con->server.addr_len = address_sa_len(result.address);
         assert(con->server.addr_len <= sizeof(con->server.addr));
-        memcpy(&con->server.addr, address_sa(server_address),
+        memcpy(&con->server.addr, address_sa(result.address),
             con->server.addr_len);
+        con->use_proxy_header = result.use_proxy_header;
 
-        free(server_address);
+        if (result.caller_free_address)
+            free((void *)result.address);
 
         con->state = RESOLVED;
     } else {
@@ -498,7 +600,8 @@ resolv_cb(struct Address *result, void *data) {
 
 static void
 free_resolv_cb_data(struct resolv_cb_data *cb_data) {
-    free(cb_data->address);
+    if (cb_data->cb_free_addr)
+        free((void *)cb_data->address);
     free(cb_data);
 }
 
@@ -558,7 +661,6 @@ initiate_server_connect(struct Connection *con, struct ev_loop *loop) {
             return;
         }
 
-        result = 0;
         int tries = 5;
         do {
             result = bind(sockfd,
@@ -590,6 +692,21 @@ initiate_server_connect(struct Connection *con, struct ev_loop *loop) {
         return;
     }
 
+    if (getsockname(sockfd, (struct sockaddr *)&con->server.local_addr,
+                &con->server.local_addr_len) != 0) {
+        close(sockfd);
+        warn("getsockname failed: %s", strerror(errno));
+
+        abort_connection(con);
+        return;
+    }
+
+    if (con->header_len && !con->use_proxy_header) {
+        /* If we prepended the PROXY header and this backend isn't configured
+         * to receive it, consume it now */
+        buffer_pop(con->client.buffer, NULL, con->header_len);
+    }
+
     struct ev_io *server_watcher = &con->server.watcher;
     ev_io_init(server_watcher, connection_cb, sockfd, EV_WRITE);
     con->server.watcher.data = con;
@@ -683,10 +800,16 @@ new_connection(struct ev_loop *loop) {
 
     con->state = NEW;
     con->client.addr_len = sizeof(con->client.addr);
+    con->client.local_addr = (struct sockaddr_storage){.ss_family = AF_UNSPEC};
+    con->client.local_addr_len = sizeof(con->client.local_addr);
     con->server.addr_len = sizeof(con->server.addr);
+    con->server.local_addr = (struct sockaddr_storage){.ss_family = AF_UNSPEC};
+    con->server.local_addr_len = sizeof(con->server.local_addr);
     con->hostname = NULL;
     con->hostname_len = 0;
+    con->header_len = 0;
     con->query_handle = NULL;
+    con->use_proxy_header = 0;
 
     con->client.buffer = new_buffer(4096, loop);
     if (con->client.buffer == NULL) {
@@ -716,12 +839,12 @@ log_connection(struct Connection *con) {
         duration = con->server.buffer->last_recv - con->established_timestamp;
 
     display_sockaddr(&con->client.addr, client_address, sizeof(client_address));
-    display_address(con->listener->address, listener_address, sizeof(listener_address));
+    display_sockaddr(&con->client.local_addr, listener_address, sizeof(listener_address));
     display_sockaddr(&con->server.addr, server_address, sizeof(server_address));
 
     log_msg(con->listener->access_log,
            LOG_NOTICE,
-           "%s -> %s -> %s [%.*s] %ld/%ld bytes tx %ld/%ld bytes rx %1.3f seconds",
+           "%s -> %s -> %s [%.*s] %zu/%zu bytes tx %zu/%zu bytes rx %1.3f seconds",
            client_address,
            listener_address,
            server_address,
@@ -737,21 +860,27 @@ log_connection(struct Connection *con) {
 static void
 log_bad_request(struct Connection *con __attribute__((unused)), const char *req, size_t req_len, int parse_result) {
     size_t message_len = 64 + 6 * req_len;
-    char *message = alloca(message_len);
+    char *message = malloc(message_len);
+    if (message == NULL) {
+        err("log_bad_request: unable to allocate message buffer");
+        return;
+    }
     char *message_pos = message;
     char *message_end = message + message_len;
 
-    message_pos += snprintf(message_pos, message_end - message_pos,
+    message_pos += snprintf(message_pos, (size_t)(message_end - message_pos),
                             "parse_packet({");
 
     for (size_t i = 0; i < req_len; i++)
-        message_pos += snprintf(message_pos, message_end - message_pos,
+        message_pos += snprintf(message_pos, (size_t)(message_end - message_pos),
                                 "0x%02hhx, ", (unsigned char)req[i]);
 
     message_pos -= 2;/* Delete the trailing ', ' */
-    message_pos += snprintf(message_pos, message_end - message_pos,
-                            "}, %ld, ...) = %d", req_len, parse_result);
+    snprintf(message_pos, (size_t)(message_end - message_pos), "}, %zu, ...) = %d",
+             req_len, parse_result);
     debug("%s", message);
+
+    free(message);
 }
 
 /*
@@ -783,39 +912,39 @@ print_connection(FILE *file, const struct Connection *con) {
         case ACCEPTED:
             fprintf(file, "ACCEPTED      %s %zu/%zu\t-\n",
                     display_sockaddr(&con->client.addr, client, sizeof(client)),
-                    con->client.buffer->len, con->client.buffer->size);
+                    buffer_len(con->client.buffer), buffer_size(con->client.buffer));
             break;
         case PARSED:
             fprintf(file, "PARSED        %s %zu/%zu\t-\n",
                     display_sockaddr(&con->client.addr, client, sizeof(client)),
-                    con->client.buffer->len, con->client.buffer->size);
+                    buffer_len(con->client.buffer), buffer_size(con->client.buffer));
             break;
         case RESOLVING:
             fprintf(file, "RESOLVING      %s %zu/%zu\t-\n",
                     display_sockaddr(&con->client.addr, client, sizeof(client)),
-                    con->client.buffer->len, con->client.buffer->size);
+                    buffer_len(con->client.buffer), buffer_size(con->client.buffer));
             break;
         case RESOLVED:
             fprintf(file, "RESOLVED      %s %zu/%zu\t-\n",
                     display_sockaddr(&con->client.addr, client, sizeof(client)),
-                    con->client.buffer->len, con->client.buffer->size);
+                    buffer_len(con->client.buffer), buffer_size(con->client.buffer));
             break;
         case CONNECTED:
             fprintf(file, "CONNECTED     %s %zu/%zu\t%s %zu/%zu\n",
                     display_sockaddr(&con->client.addr, client, sizeof(client)),
-                    con->client.buffer->len, con->client.buffer->size,
+                    buffer_len(con->client.buffer), buffer_size(con->client.buffer),
                     display_sockaddr(&con->server.addr, server, sizeof(server)),
-                    con->server.buffer->len, con->server.buffer->size);
+                    buffer_len(con->server.buffer), buffer_size(con->server.buffer));
             break;
         case SERVER_CLOSED:
             fprintf(file, "SERVER_CLOSED %s %zu/%zu\t-\n",
                     display_sockaddr(&con->client.addr, client, sizeof(client)),
-                    con->client.buffer->len, con->client.buffer->size);
+                    buffer_len(con->client.buffer), buffer_size(con->client.buffer));
             break;
         case CLIENT_CLOSED:
             fprintf(file, "CLIENT_CLOSED -\t%s %zu/%zu\n",
                     display_sockaddr(&con->server.addr, server, sizeof(server)),
-                    con->server.buffer->len, con->server.buffer->size);
+                    buffer_len(con->server.buffer), buffer_size(con->server.buffer));
             break;
         case CLOSED:
             fprintf(file, "CLOSED        -\t-\n");
